from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, ParamSpec, cast
from uuid import uuid4

from typing_extensions import Self, TypeVar

from eventsourcing.domain import (
    AbstractDecision,
    CallableType,
    ProgrammingError,
    all_func_decorators,
    decorated_func_callers,
    filter_kwargs_for_method_params,
)
from eventsourcing.utils import construct_topic, get_topic, resolve_topic

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

_enduring_object_init_classes: dict[type[Any], type[InitialDecision]] = {}


class Decision(AbstractDecision):
    def as_dict(self) -> dict[str, Any]:
        return self.__dict__.copy()

    def mutate(self, obj: TPerspective | None) -> TPerspective | None:
        assert obj is not None

        # Identify the function that was decorated.
        try:
            decorated_func = decorated_funcs[(type(obj), type(self))]
        except KeyError:
            pass
        else:
            # Select event attributes mentioned in function signature.
            self_dict = self.as_dict()
            kwargs = filter_kwargs_for_method_params(self_dict, decorated_func)

            # Call the original method with event attribute values.
            decorated_method = decorated_func.__get__(obj, type(obj))
            try:
                decorated_method(**kwargs)
            except TypeError as e:  # pragma: no cover
                # TODO: Write a test that does this...
                msg = (
                    f"Failed to apply {type(self).__qualname__} to "
                    f"{type(obj).__qualname__} with kwargs {kwargs}: {e}"
                )
                raise TypeError(msg) from e

        self.apply(obj)
        return obj

    def apply(self, obj: Any) -> None:
        pass


class InitialDecision(Decision):
    originator_topic: str

    def mutate(self, obj: TPerspective | None) -> TPerspective | None:
        # Identify the function that was decorated.
        if obj is not None:
            try:
                decorated_func = decorated_funcs[(type(obj), type(self))]
            except KeyError:  # pragma: no cover
                pass
            else:
                # Select event attributes mentioned in function signature.
                self_dict = self.as_dict()
                kwargs = filter_kwargs_for_method_params(self_dict, decorated_func)

                # Call the original method with event attribute values.
                decorated_method = decorated_func.__get__(obj, type(obj))
                try:
                    decorated_method(**kwargs)
                except TypeError as e:  # pragma: no cover
                    # TODO: Write a test that does this...
                    msg = (
                        f"Failed to apply {type(self).__qualname__} to "
                        f"{type(obj).__qualname__} with kwargs {kwargs}: {e}"
                    )
                    raise TypeError(msg) from e
                return obj

        kwargs = self.as_dict()
        originator_type = resolve_topic(kwargs.pop("originator_topic"))
        if issubclass(originator_type, EnduringObject):
            enduring_object_id = kwargs.pop(self.id_attr_name(originator_type))
            try:
                enduring_object = type.__call__(originator_type, **kwargs)
            except TypeError as e:  # pragma: no cover
                msg = (
                    f"{type(self).__qualname__} cannot __init__ "
                    f"{originator_type.__qualname__} "
                    f"with kwargs {kwargs}: {e}"
                )
                raise TypeError(msg) from e
            enduring_object.id = enduring_object_id
            return enduring_object
        msg = f"Originator type not subclass of EnduringObject: {originator_type}"
        raise TypeError(msg)

    @classmethod
    def id_attr_name(cls, enduring_object_class: type[EnduringObject[Any, TID]]) -> TID:
        return cast(TID, f"{enduring_object_class.__name__.lower()}_id")


TDecision = TypeVar("TDecision", bound=Decision)
"""
A type variable representing any subclass of :class:`Decision`.
"""


class Tagged(Generic[TDecision]):
    def __init__(self, tags: list[str], decision: TDecision) -> None:
        self.tags = tags
        self.decision = decision


T = TypeVar("T")
P = ParamSpec("P")


class MetaPerspective(ABCMeta):
    pass


class Perspective(ABC, Generic[TDecision], metaclass=MetaPerspective):
    last_known_position: int | None
    new_decisions: list[Tagged[TDecision]]

    def __new__(cls, *_: Any, **__: Any) -> Self:
        self = super().__new__(cls)
        self.last_known_position = None
        self.new_decisions = []
        return self

    @abstractmethod
    def consistency_boundary(self) -> Selector | Sequence[Selector]:
        raise NotImplementedError  # pragma: no cover

    def trigger_event(
        self,
        decision_cls: Callable[P, TDecision],
        tags: Sequence[str] = (),
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        Constructs new tagged decision and appends to list of uncommitted events.
        """
        tagged = Tagged[TDecision](
            tags=list(tags),
            decision=decision_cls(*args, **kwargs),
        )
        tagged.decision.mutate(self)
        self.new_decisions.append(tagged)

    def collect_events(self) -> Sequence[Tagged[TDecision]]:
        """
        Drains list of triggered events.
        """
        collected, self.new_decisions = self.new_decisions, []
        return collected


TPerspective = TypeVar("TPerspective", bound=Perspective[Any])


decorated_funcs: dict[tuple[MetaPerspective, type[Decision]], CallableType] = {}


class MetaSupportsEventDecorator(MetaPerspective):
    def __init__(
        cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)

        topic_prefix = construct_topic(cls) + "."

        cls.projected_types: list[type[Decision]] = []

        # Find the event decorators on this class.
        func_decorators = [
            decorator
            for decorator in all_func_decorators
            if construct_topic(decorator.decorated_func).startswith(topic_prefix)
        ]

        for decorator in func_decorators:
            given = decorator.given_event_cls

            # Keep things simple by only supporting given classes (not names).
            assert given is not None, "Event class not given"
            # TODO: Maybe support event name strings, maybe not....

            # Make sure given event class is a Decision subclass.
            assert issubclass(given, Decision)

            if (
                issubclass(given, InitialDecision)
                and decorator.decorated_func.__name__ == "__init__"
            ):
                _enduring_object_init_classes[cls] = given
                # If command method, remember which event class to trigger.
            elif not construct_topic(decorator.decorated_func).endswith("._"):
                decorated_func_callers[decorator] = given

            # Remember which decorated func to call.
            decorated_funcs[(cls, given)] = decorator.decorated_func

            cls.projected_types.append(given)


class MetaEnduringObject(MetaSupportsEventDecorator):
    def __init__(
        cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)
        # Find and remember the "InitialDecision" class.
        for item in cls.__dict__.values():
            if isinstance(item, type) and issubclass(item, InitialDecision):
                _enduring_object_init_classes[cls] = item
                break

    def __call__(cls: type[T], **kwargs: Any) -> T:
        # TODO: For convenience, make this error out in the same way
        #  as it would if the arguments didn't match the __init__
        #  method and __init__was called directly, and verify the
        #  event's __init__ is valid when initialising the class,
        #  just like we do for event-sourced aggregates.

        assert issubclass(cls, EnduringObject)
        try:
            init_enduring_object_class = _enduring_object_init_classes[cls]
        except KeyError:
            msg = (
                f"Enduring object class {cls.__name__} has no "
                f"InitialDecision class. Please define a subclass of "
                f"InitialDecision as a nested class on {cls.__name__}."
            )
            raise ProgrammingError(msg) from None

        return cls._create(
            decision_cls=init_enduring_object_class,
            **kwargs,
        )


TID = TypeVar("TID", bound=str, default=str)


class EnduringObject(
    Perspective[TDecision], Generic[TDecision, TID], metaclass=MetaEnduringObject
):
    id: TID

    @classmethod
    def _create(
        cls: type[Self], decision_cls: type[InitialDecision], **kwargs: Any
    ) -> Self:
        enduring_object_id = cls._create_id()
        id_attr_name = decision_cls.id_attr_name(cls)
        assert id_attr_name not in kwargs
        assert "originator_topic" not in kwargs
        assert "tags" not in kwargs
        initial_kwargs: dict[str, Any] = {
            id_attr_name: enduring_object_id,
            "originator_topic": get_topic(cls),
        }
        initial_kwargs.update(kwargs)
        try:

            tagged = Tagged[TDecision](
                tags=[enduring_object_id],
                decision=cast(type[TDecision], decision_cls)(**initial_kwargs),
            )
        except TypeError as e:
            msg = (
                f"Unable to construct {decision_cls.__qualname__} event "
                f"with kwargs {initial_kwargs}: {e}"
            )
            raise TypeError(msg) from e
        self = cast(Self, tagged.decision.mutate(None))
        assert self is not None
        self.new_decisions.append(tagged)
        return self

    @classmethod
    def _create_id(cls) -> TID:
        return cast(TID, f"{cls.__name__.lower()}-{uuid4()}")

    def consistency_boundary(self) -> list[Selector]:
        return [Selector(tags=[self.id])]

    def trigger_event(
        self,
        decision_cls: Callable[P, TDecision],
        tags: Sequence[str] = (),
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        tags = [self.id, *tags]
        super().trigger_event(decision_cls, tags, *args, **kwargs)


class Group(Perspective[TDecision]):
    _enduring_objects: list[EnduringObject[TDecision]]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self = super().__new__(cls, *args, **kwargs)
        self._enduring_objects = [a for a in args if isinstance(a, EnduringObject)]
        return self

    def consistency_boundary(self) -> list[Selector]:
        return [
            Selector(tags=cb.tags)
            for cbs in [o.consistency_boundary() for o in self._enduring_objects]
            for cb in cbs
        ]

    def trigger_event(
        self,
        decision_cls: Callable[P, TDecision],
        tags: Sequence[str] = (),
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        objs = self._enduring_objects
        tags = [o.id for o in objs] + list(tags)
        tagged = Tagged[TDecision](
            tags=tags,
            decision=decision_cls(*args, **kwargs),
        )
        for o in objs:
            tagged.decision.mutate(o)
        self.new_decisions.append(tagged)


@dataclass
class Selector:
    types: Sequence[type[Decision]] = ()
    tags: Sequence[str] = ()


class MetaSlice(MetaSupportsEventDecorator):
    def __init__(
        cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> None:
        super().__init__(name, bases, namespace)
        cls.do_projection = len(cls.projected_types) != 0


class Slice(Perspective[TDecision], metaclass=MetaSlice):
    def execute(self) -> None:
        pass


TSlice = TypeVar("TSlice", bound=Slice[Any])
TGroup = TypeVar("TGroup", bound=Group[Any])



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import ClassVar
from unittest import TestCase
from uuid import NAMESPACE_URL, UUID, uuid5

from eventsourcing.application import (
    AggregateNotFoundError,
    Application,
    ProcessingEvent,
)
from eventsourcing.dcb.application import DCBApplication
from eventsourcing.dcb.domain import EnduringObject, Tagged
from eventsourcing.dcb.msgpack import Decision, InitialDecision
from eventsourcing.dispatch import singledispatchmethod
from eventsourcing.domain import Aggregate, DomainEventProtocol, event
from eventsourcing.persistence import (
    IntegrityError,
    Tracking,
    TrackingRecorder,
)
from eventsourcing.projection import (
    EventSourcedProjection,
    EventSourcedProjectionRunner,
    Projection,
    ProjectionRunner,
)
from eventsourcing.utils import get_topic


class Counter(Aggregate):
    def __init__(self, name: str) -> None:
        self.name = name
        self.count = 0

    @classmethod
    def create_id(cls, name: str) -> UUID:
        return uuid5(NAMESPACE_URL, f"/counters/{name}")

    @event("Incremented")
    def increment(self) -> None:
        self.count += 1


class EventCountersInterface(TrackingRecorder):
    @abstractmethod
    def get_created_event_counter(self) -> int:
        pass

    @abstractmethod
    def get_subsequent_event_counter(self) -> int:
        pass

    @abstractmethod
    def incr_created_event_counter(self, tracking: Tracking) -> None:
        pass

    @abstractmethod
    def incr_subsequent_event_counter(self, tracking: Tracking) -> None:
        pass


class Counters(EventSourcedProjection[UUID]):
    @singledispatchmethod
    def policy(
        self,
        domain_event: DomainEventProtocol[UUID],
        processing_event: ProcessingEvent[UUID],
    ) -> None:
        topic = get_topic(type(domain_event))
        try:
            counter_id = Counter.create_id(topic)
            counter: Counter = self.repository.get(counter_id)
        except AggregateNotFoundError:
            counter = Counter(topic)
        counter.increment()
        processing_event.collect_events(counter)

    def get_count(self, domain_event_class: type[DomainEventProtocol[UUID]]) -> int:
        topic = get_topic(domain_event_class)
        counter_id = Counter.create_id(topic)
        try:
            counter: Counter = self.repository.get(counter_id)
        except AggregateNotFoundError:
            return 0
        return counter.count


class EventCountersViewTestCase(TestCase):
    def construct_event_counters_view(self) -> EventCountersInterface:
        raise NotImplementedError

    def test(self) -> None:
        # Construct materialised view object.
        view = self.construct_event_counters_view()

        # Check the view object is a tracking recorder.
        self.assertIsInstance(view, TrackingRecorder)

        # Check the view has processed no events.
        self.assertIsNone(view.max_tracking_id("upstream"))

        # Check the event counters are zero.
        self.assertEqual(view.get_created_event_counter(), 0)
        self.assertEqual(view.get_subsequent_event_counter(), 0)

        # Increment the "created" event counter.
        view.incr_created_event_counter(Tracking("upstream", 1))

        # Check the counted number of events.
        self.assertEqual(view.get_created_event_counter(), 1)
        self.assertEqual(view.get_subsequent_event_counter(), 0)

        # Increment the subsequent event counter.
        view.incr_subsequent_event_counter(Tracking("upstream", 2))

        # Check the counted number of events.
        self.assertEqual(view.get_created_event_counter(), 1)
        self.assertEqual(view.get_subsequent_event_counter(), 1)

        # Increment the subsequent event counter again.
        view.incr_subsequent_event_counter(Tracking("upstream", 3))

        # Check the counted number of events.
        self.assertEqual(view.get_created_event_counter(), 1)
        self.assertEqual(view.get_subsequent_event_counter(), 2)

        # Check the tracking objects have been recorded.
        self.assertEqual(view.max_tracking_id("upstream"), 3)

        # Check the tracking objects are recorded uniquely and atomically.
        with self.assertRaises(IntegrityError):
            view.incr_created_event_counter(Tracking("upstream", 3))

        # Check the counted number of events.
        self.assertEqual(view.get_created_event_counter(), 1)

        with self.assertRaises(IntegrityError):
            view.incr_subsequent_event_counter(Tracking("upstream", 3))

        # Check the counted number of events.
        self.assertEqual(view.get_subsequent_event_counter(), 2)

        # Check the wait() method returns normally when
        # we wait for a position that has been recorded.
        view.wait("upstream", 3)

        # Check the wait() method raises a TimeoutError when
        # we wait for a postion that has not been recorded.
        with self.assertRaises(TimeoutError):
            view.wait("upstream", 4, timeout=0.5)


class SpannerThrown(Aggregate.Event):
    pass


class SpannerThrownError(Exception):
    pass


class DCBSpannerThrown(Decision):
    # Avoid segmentation violation with Python 3.13
    # and MsgStruct instances with zero attributes.
    a: str


# Define a perspective.
class Thing(EnduringObject[Decision, str]):
    class Created(InitialDecision):
        thing_id: str


class TaggedDecisionCountersProjection(Projection[EventCountersInterface]):
    name = "eventcounters"
    topics: tuple[str, ...] = (
        get_topic(Thing.Created),
        get_topic(Decision),
        get_topic(DCBSpannerThrown),
    )

    @singledispatchmethod
    def process_event(self, tagged: Tagged[Decision], tracking: Tracking) -> None:
        self.process_decision(tagged.decision, tracking)

    @singledispatchmethod
    def process_decision(self, _: Decision, tracking: Tracking) -> None:
        self.view.insert_tracking(tracking)

    @process_decision.register
    def _(self, _: InitialDecision, tracking: Tracking) -> None:
        self.view.incr_created_event_counter(tracking)

    @process_decision.register
    def _(self, _: Decision, tracking: Tracking) -> None:
        self.view.incr_subsequent_event_counter(tracking)

    @process_decision.register
    def _(self, _: DCBSpannerThrown, __: Tracking) -> None:
        msg = "This is a deliberate bug"
        raise SpannerThrownError(msg)


class AggregateEventCountersProjection(Projection[EventCountersInterface]):
    name = "eventcounters"
    topics: tuple[str, ...] = (
        get_topic(Aggregate.Created),
        get_topic(Aggregate.Event),
        get_topic(SpannerThrown),
    )

    @singledispatchmethod
    def process_event(self, _: Aggregate.Event, tracking: Tracking) -> None:
        self.view.insert_tracking(tracking)

    @process_event.register
    def aggregate_created(self, _: Aggregate.Created, tracking: Tracking) -> None:
        self.view.incr_created_event_counter(tracking)

    @process_event.register
    def aggregate_event(self, _: Aggregate.Event, tracking: Tracking) -> None:
        self.view.incr_subsequent_event_counter(tracking)

    @process_event.register
    def spanner_thrown(self, _: SpannerThrown, __: Tracking) -> None:
        msg = "This is a deliberate bug"
        raise SpannerThrownError(msg)


class AggregateEventCountersProjectionTestCase(TestCase, ABC):
    view_class: type[EventCountersInterface]
    env: ClassVar[dict[str, str]] = {}

    def test_event_counters_projection(self) -> None:
        # Construct runner with application, projection, and recorder.
        runner = ProjectionRunner(
            application_class=Application[UUID],
            projection_class=AggregateEventCountersProjection,
            view_class=self.view_class,
            env=self.env,
        )
        with runner:

            # Get "read" and "write" model instances from the runner.
            write_model = runner.app
            read_model = runner.projection.view

            # Write some events.
            aggregate = Aggregate()
            aggregate.trigger_event(event_class=Aggregate.Event)
            aggregate.trigger_event(event_class=Aggregate.Event)
            recordings = write_model.save(aggregate)

            # Wait for the events to be processed.
            read_model.wait(
                application_name=write_model.name,
                notification_id=recordings[-1].notification.id,
            )

            # Query the read model.
            self.assertEqual(read_model.get_created_event_counter(), 1)
            self.assertEqual(read_model.get_subsequent_event_counter(), 2)

            # Write some more events.
            aggregate = Aggregate()
            aggregate.trigger_event(event_class=Aggregate.Event)
            aggregate.trigger_event(event_class=Aggregate.Event)
            recordings = write_model.save(aggregate)

            # Wait for the events to be processed.
            read_model.wait(
                application_name=write_model.name,
                notification_id=recordings[-1].notification.id,
            )

            # Query the read model.
            self.assertEqual(read_model.get_created_event_counter(), 2)
            self.assertEqual(read_model.get_subsequent_event_counter(), 4)

    def test_run_forever_raises_projection_error(self) -> None:
        # Construct runner with application, projection, and recorder.
        with ProjectionRunner(
            application_class=Application[UUID],
            projection_class=AggregateEventCountersProjection,
            view_class=self.view_class,
            env=self.env,
        ) as runner:
            write_model = runner.app
            read_model = runner.projection.view

            # Write some events.
            aggregate = Aggregate()
            aggregate.trigger_event(event_class=SpannerThrown)
            recordings = write_model.save(aggregate)

            # Projection runner terminates with projection error.
            with self.assertRaises(SpannerThrownError):
                runner.run_forever(timeout=5)

            # Wait times out (event has not been processed).
            with self.assertRaises(TimeoutError):
                read_model.wait(
                    application_name=write_model.name,
                    notification_id=recordings[-1].notification.id,
                )


class TaggedDecisionCountersProjectionTestCase(TestCase, ABC):
    view_class: type[EventCountersInterface]
    env: ClassVar[dict[str, str]]

    def test_event_counters_projection(self) -> None:

        # Construct runner with application, projection, and recorder.
        with ProjectionRunner(
            application_class=DCBApplication,
            projection_class=TaggedDecisionCountersProjection,
            view_class=self.view_class,
            env=self.env,
        ) as runner:

            # Get "read" and "write" model instances from the runner.
            write_model = runner.app
            read_model = runner.projection.view

            # Write some events.
            perspective = Thing()
            perspective.trigger_event(Decision)
            perspective.trigger_event(Decision)
            self.assertEqual(3, len(perspective.new_decisions))
            position = write_model.repository.save(perspective)

            # Wait for the events to be processed.
            read_model.wait(
                application_name=write_model.name,
                notification_id=position,
            )

            # Query the read model.
            self.assertEqual(read_model.get_created_event_counter(), 1)
            self.assertEqual(read_model.get_subsequent_event_counter(), 2)

            # Write some more events.
            perspective = Thing()
            perspective.trigger_event(Decision)
            perspective.trigger_event(Decision)
            position = write_model.repository.save(perspective)

            # Wait for the events to be processed.
            read_model.wait(
                application_name=write_model.name,
                notification_id=position,
            )

            # Query the read model.
            self.assertEqual(read_model.get_created_event_counter(), 2)
            self.assertEqual(read_model.get_subsequent_event_counter(), 4)

    def test_run_forever_raises_projection_error(self) -> None:
        # Construct runner with application, projection, and recorder.
        with ProjectionRunner(
            application_class=DCBApplication,
            projection_class=TaggedDecisionCountersProjection,
            view_class=self.view_class,
            env=self.env,
        ) as runner:
            write_model = runner.app
            read_model = runner.projection.view

            # Write some events.
            perspective = Thing()
            perspective.trigger_event(DCBSpannerThrown, a="")
            position = write_model.repository.save(perspective)

            # Projection runner terminates with projection error.
            with self.assertRaises(SpannerThrownError):
                runner.run_forever(timeout=5)

            # Wait times out (event has not been processed).
            with self.assertRaises(TimeoutError):
                read_model.wait(
                    application_name=write_model.name,
                    notification_id=position,
                )


class EventSourcedProjectionTestCase(TestCase):
    env: ClassVar[dict[str, str]] = {}

    def test_event_sourced_projection(self) -> None:
        with EventSourcedProjectionRunner(
            application_class=Application, projection_class=Counters, env=self.env
        ) as runner:
            recordings = runner.app.save(Aggregate())
            runner.wait(recordings[-1].notification.id)
            self.assertEqual(1, runner.projection.get_count(Aggregate.Created))
            self.assertEqual(0, runner.projection.get_count(Aggregate.Event))

            recordings = runner.app.save(Aggregate())
            runner.wait(recordings[-1].notification.id)
            self.assertEqual(2, runner.projection.get_count(Aggregate.Created))
            self.assertEqual(0, runner.projection.get_count(Aggregate.Event))

            recordings = runner.app.save(Aggregate())
            runner.wait(recordings[-1].notification.id)
            self.assertEqual(3, runner.projection.get_count(Aggregate.Created))
            self.assertEqual(0, runner.projection.get_count(Aggregate.Event))

            aggregate = Aggregate()
            aggregate.trigger_event(Aggregate.Event)
            recordings = runner.app.save(aggregate)
            runner.wait(recordings[-1].notification.id)
            self.assertEqual(4, runner.projection.get_count(Aggregate.Created))
            self.assertEqual(1, runner.projection.get_count(Aggregate.Event))



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from eventsourcing.persistence import ProgrammingError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typing_extensions import Self


@dataclass
class DCBQueryItem:
    types: list[str] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)


@dataclass
class DCBQuery:
    items: list[DCBQueryItem] = field(default_factory=list)


@dataclass
class DCBAppendCondition:
    fail_if_events_match: DCBQuery = field(default_factory=DCBQuery)
    after: int | None = None


@dataclass
class DCBEvent:
    type: str
    data: bytes
    tags: list[str] = field(default_factory=list)


@dataclass
class DCBSequencedEvent:
    event: DCBEvent
    position: int


class DCBReadResponse(Iterator[DCBSequencedEvent], ABC):
    @property
    @abstractmethod
    def head(self) -> int | None:
        pass  # pragma: no cover

    @abstractmethod
    def __next__(self) -> DCBSequencedEvent:
        pass  # pragma: no cover

    # @abstractmethod
    # def next_batch(self) -> list[DCBSequencedEvent]:
    #     """
    #     Returns a batch of events as a list.
    #     Updates the head position similar to __next__.
    #     """


class DCBRecorder(ABC):
    @abstractmethod
    def read(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
        limit: int | None = None,
    ) -> DCBReadResponse:
        """
        Returns all events, unless 'after' is given then only those with position
        greater than 'after', and unless any query items are given, then only those
        that match at least one query item. An event matches a query item if its type
        is in the item types or there are no item types, and if all the item tags are
        in the event tags.
        """

    @abstractmethod
    def append(
        self, events: Sequence[DCBEvent], condition: DCBAppendCondition | None = None
    ) -> int:
        """
        Appends given events to the event store, unless the condition fails.
        """

    @abstractmethod
    def subscribe(
        self,
        query: DCBQuery | None = None,
        *,
        after: int | None = None,
    ) -> DCBSubscription[Self]:
        """
        Returns all events, unless 'after' is given then only those with position
        greater than 'after', and unless any query items are given, then only those
        that match at least one query item. An event matches a query item if its type
        is in the item types or there are no item types, and if all the item tags are
        in the event tags. The subscription will block when the last recorded event
        is received, and then continue when new events are recorded.
        """


TDCBRecorder_co = TypeVar("TDCBRecorder_co", bound=DCBRecorder, covariant=True)


class DCBSubscription(Iterator[DCBSequencedEvent], Generic[TDCBRecorder_co]):
    def __init__(
        self,
        recorder: TDCBRecorder_co,
        query: DCBQuery | None = None,
        after: int | None = None,
    ) -> None:
        self._recorder = recorder
        self._query = query
        self._has_been_entered = False
        self._has_been_stopped = False
        self._last_position: int = after or 0

    def __enter__(self) -> Self:
        if self._has_been_entered:
            msg = "Already entered subscription context manager"
            raise ProgrammingError(msg)
        self._has_been_entered = True
        return self

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        if not self._has_been_entered:
            msg = "Not already entered subscription context manager"
            raise ProgrammingError(msg)
        self.stop()

    def stop(self) -> None:
        """Stops the subscription."""
        self._has_been_stopped = True

    def __iter__(self) -> Self:
        return self

    @abstractmethod
    def __next__(self) -> DCBSequencedEvent:
        """Returns the next DCBEvent in the sequence."""



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Generic, cast

from eventsourcing.dcb.domain import (
    EnduringObject,
    InitialDecision,
    Perspective,
    Selector,
    TDecision,
    TGroup,
    TPerspective,
    TSlice,
)
from eventsourcing.dcb.persistence import (
    DCBEventStore,
    DCBInfrastructureFactory,
    DCBMapper,
    NotFoundError,
)
from eventsourcing.utils import Environment, EnvType, resolve_topic

if TYPE_CHECKING:
    from collections.abc import Mapping
    from types import TracebackType

    from typing_extensions import Self


class DCBApplication:
    name = "DCBApplication"
    env: Mapping[str, str] = {"PERSISTENCE_MODULE": "eventsourcing.dcb.popo"}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if "name" not in cls.__dict__:
            cls.name = cls.__name__

    def __init__(self, env: EnvType | None = None):
        self.env = self.construct_env(self.name, env)
        self.factory = DCBInfrastructureFactory.construct(self.env)
        self.recorder = self.factory.dcb_recorder()
        if "MAPPER_TOPIC" in self.env:
            # Only need a mapper, event store, and repository
            # if we are using the higher-level abstractions.
            self.mapper = cast(
                DCBMapper[Any], resolve_topic(self.env["MAPPER_TOPIC"])()
            )
            assert isinstance(self.mapper, DCBMapper)
            self.events = DCBEventStore(self.mapper, self.recorder)
            self.repository = DCBRepository(self.events)

    def construct_env(self, name: str, env: EnvType | None = None) -> Environment:
        """Constructs environment from which application will be configured."""
        _env = dict(type(self).env)
        _env.update(os.environ)
        if env is not None:
            _env.update(env)
        return Environment(name, _env)

    def do(self, s: TSlice) -> TSlice:
        """
        Advances and executes a slice, then saves new decisions.
        """
        if type(s).do_projection:
            s = self.repository.advance(s)
        s.execute()
        if s.new_decisions:
            self.repository.save(s)
        return s

    def close(self) -> None:
        self.factory.close()

    def __enter__(self) -> Self:
        self.factory.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()
        self.factory.__exit__(exc_type, exc_val, exc_tb)


class DCBRepository(Generic[TDecision]):
    def __init__(self, eventstore: DCBEventStore[TDecision]):
        self.eventstore = eventstore

    def save(self, p: Perspective[TDecision]) -> int:
        return self.eventstore.append(
            events=p.collect_events(),
            cb=p.consistency_boundary(),
            after=p.last_known_position,
        )

    def get(
        self,
        enduring_object_id: str,
    ) -> EnduringObject[TDecision]:
        cb = [Selector(tags=[enduring_object_id])]
        events = self.eventstore.read(*cb)
        obj: EnduringObject[TDecision] | None = None
        for event in events:
            obj = event.decision.mutate(obj)
        if obj is None:
            raise NotFoundError
        obj.last_known_position = events.head
        return obj

    def get_many(
        self,
        *enduring_object_ids: str,
    ) -> list[EnduringObject[TDecision] | None]:
        cb = [
            Selector(tags=[enduring_object_id])
            for enduring_object_id in enduring_object_ids
        ]
        tagged_decisions = self.eventstore.read(cb)
        objs: dict[str, EnduringObject[TDecision] | None] = dict.fromkeys(
            enduring_object_ids
        )
        for tagged in tagged_decisions:
            for tag in tagged.tags:
                obj = objs.get(tag)
                if not isinstance(tagged.decision, InitialDecision) and not obj:
                    continue
                obj = tagged.decision.mutate(obj)
                objs[tag] = obj
        for obj in objs.values():
            if obj is not None:
                obj.last_known_position = tagged_decisions.head
        return list(objs.values())

    def get_group(self, cls: type[TGroup], *enduring_object_ids: str) -> TGroup:
        enduring_objects = self.get_many(*enduring_object_ids)
        perspective = cls(*enduring_objects)
        last_known_positions = [
            o.last_known_position
            for o in enduring_objects
            if o and o.last_known_position
        ]
        perspective.last_known_position = (
            max(last_known_positions) if last_known_positions else None
        )
        return perspective

    def advance(self, p: TPerspective) -> TPerspective:
        events = self.eventstore.read(
            cb=p.consistency_boundary(),
            after=p.last_known_position,
        )
        for event in events:
            event.decision.mutate(p)
        p.last_known_position = events.head
        return p



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from base64 import b64decode, b64encode
from typing import TYPE_CHECKING, Generic
from uuid import UUID

from eventsourcing.application import NotificationLog, Section, TApplication
from eventsourcing.persistence import Notification

if TYPE_CHECKING:
    from collections.abc import Sequence


class NotificationLogInterface(ABC):
    """Abstract base class for obtaining serialised
    sections of a notification log.
    """

    @abstractmethod
    def get_log_section(self, section_id: str) -> str:
        """Returns a serialised :class:`~eventsourcing.application.Section`
        from a notification log.
        """

    @abstractmethod
    def get_notifications(
        self,
        start: int | None,
        limit: int,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> str:
        """Returns a serialised list of :class:`~eventsourcing.persistence.Notification`
        objects from a notification log.
        """


class NotificationLogJSONService(NotificationLogInterface, Generic[TApplication]):
    """Presents serialised sections of a notification log."""

    def __init__(self, app: TApplication):
        """Initialises service with given application."""
        self.app = app

    def get_log_section(self, section_id: str) -> str:
        """Returns JSON serialised :class:`~eventsourcing.application.Section`
        from a notification log.
        """
        section = self.app.notification_log[section_id]
        return json.dumps(
            {
                "id": section.id,
                "next_id": section.next_id,
                "items": [
                    {
                        "id": item.id,
                        "originator_id": (
                            item.originator_id.hex
                            if isinstance(item.originator_id, UUID)
                            else item.originator_id
                        ),
                        "originator_version": item.originator_version,
                        "topic": item.topic,
                        "state": b64encode(item.state).decode("utf8"),
                    }
                    for item in section.items
                ],
            }
        )

    def get_notifications(
        self,
        start: int | None,
        limit: int,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> str:
        notifications = self.app.notification_log.select(
            start=start,
            limit=limit,
            topics=topics,
            inclusive_of_start=inclusive_of_start,
        )
        return json.dumps(
            [
                {
                    "id": notification.id,
                    "originator_id": (
                        notification.originator_id.hex
                        if isinstance(notification.originator_id, UUID)
                        else notification.originator_id
                    ),
                    "originator_version": notification.originator_version,
                    "topic": notification.topic,
                    "state": b64encode(notification.state).decode("utf8"),
                }
                for notification in notifications
            ]
        )


class NotificationLogJSONClient(NotificationLog):
    """Presents deserialized sections of a notification log."""

    def __init__(self, interface: NotificationLogInterface):
        """Initialises log with a given interface."""
        self.interface = interface

    def __getitem__(self, section_id: str) -> Section:
        """Returns a :class:`Section` of
        :class:`~eventsourcing.persistence.Notification` objects
        from the notification log.
        """
        body = self.interface.get_log_section(section_id)
        section = json.loads(body)
        return Section(
            id=section["id"],
            next_id=section["next_id"],
            items=[
                Notification(
                    id=item["id"],
                    originator_id=UUID(item["originator_id"]),
                    originator_version=item["originator_version"],
                    topic=item["topic"],
                    state=b64decode(item["state"].encode("utf8")),
                )
                for item in section["items"]
            ],
        )

    def select(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> list[Notification]:
        """Returns a selection of
        :class:`~eventsourcing.persistence.Notification` objects
        from the notification log.
        """
        return [
            Notification(
                id=item["id"],
                originator_id=UUID(item["originator_id"]),
                originator_version=item["originator_version"],
                topic=item["topic"],
                state=b64decode(item["state"].encode("utf8")),
            )
            for item in json.loads(
                self.interface.get_notifications(
                    start=start,
                    limit=limit,
                    topics=topics,
                    inclusive_of_start=inclusive_of_start,
                )
            )
        ]



from __future__ import annotations

import os
from base64 import b64decode, b64encode
from typing import TYPE_CHECKING

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from eventsourcing.persistence import Cipher

if TYPE_CHECKING:
    from eventsourcing.utils import Environment


class AESCipher(Cipher):
    """Cipher strategy that uses AES cipher (in GCM mode)
    from the Python cryptography package.
    """

    CIPHER_KEY = "CIPHER_KEY"
    KEY_SIZES = (16, 24, 32)

    @staticmethod
    def create_key(num_bytes: int) -> str:
        """Creates AES cipher key, with length num_bytes.

        :param num_bytes: An int value, either 16, 24, or 32.

        """
        AESCipher.check_key_size(num_bytes)
        key = AESGCM.generate_key(num_bytes * 8)
        return b64encode(key).decode("utf8")

    @staticmethod
    def check_key_size(num_bytes: int) -> None:
        if num_bytes not in AESCipher.KEY_SIZES:
            msg = f"Invalid key size: {num_bytes} not in {AESCipher.KEY_SIZES}"
            raise ValueError(msg)

    @staticmethod
    def random_bytes(num_bytes: int) -> bytes:
        return os.urandom(num_bytes)

    def __init__(self, environment: Environment):
        """Initialises AES cipher with ``cipher_key``.

        :param str cipher_key: 16, 24, or 32 bytes encoded as base64
        """
        cipher_key = environment.get(self.CIPHER_KEY)
        if not cipher_key:
            msg = f"'{self.CIPHER_KEY}' not in env"
            raise OSError(msg)
        key = b64decode(cipher_key.encode("utf8"))
        AESCipher.check_key_size(len(key))
        self.key = key

    def encrypt(self, plaintext: bytes) -> bytes:
        """Return ciphertext for given plaintext."""
        # Construct AES-GCM cipher, with 96-bit nonce.
        aesgcm = AESGCM(self.key)
        nonce = AESCipher.random_bytes(12)
        res = aesgcm.encrypt(nonce, plaintext, None)
        # Put tag at the front for compatibility with eventsourcing.crypto.AESCipher.
        tag = res[-16:]
        encrypted = res[:-16]
        return nonce + tag + encrypted

    def decrypt(self, ciphertext: bytes) -> bytes:
        """Return plaintext for given ciphertext."""
        # Split out the nonce, tag, and encrypted data.
        nonce = ciphertext[:12]
        if len(nonce) != 12:
            msg = "Damaged cipher text: invalid nonce length"
            raise ValueError(msg)

        # Expect tag at the front.
        tag = ciphertext[12:28]
        if len(tag) != 16:
            msg = "Damaged cipher text: invalid tag length"
            raise ValueError(msg)
        encrypted = ciphertext[28:]

        aesgcm = AESGCM(self.key)
        try:
            plaintext = aesgcm.decrypt(nonce, encrypted + tag, None)
        except InvalidTag as e:
            msg = "Invalid cipher tag"
            raise ValueError(msg) from e
        # Decrypt and verify.
        return plaintext



from __future__ import annotations

from typing import Any, TypeVar

import msgspec

from eventsourcing.dcb import api, domain, persistence
from eventsourcing.utils import get_topic, resolve_topic


class Decision(msgspec.Struct, domain.Decision):
    def as_dict(self) -> dict[str, Any]:
        return {key: getattr(self, key) for key in self.__struct_fields__}


TDecision = TypeVar("TDecision", bound=Decision)


class MessagePackMapper(persistence.DCBMapper[Decision]):
    def to_dcb_event(self, event: domain.Tagged[TDecision]) -> api.DCBEvent:
        return api.DCBEvent(
            type=get_topic(type(event.decision)),
            data=msgspec.msgpack.encode(event.decision),
            tags=event.tags,
        )

    def to_domain_event(self, event: api.DCBEvent) -> domain.Tagged[Decision]:
        return domain.Tagged(
            tags=event.tags,
            decision=msgspec.msgpack.decode(
                event.data,
                type=resolve_topic(event.type),
            ),
        )


class InitialDecision(Decision, domain.InitialDecision):
    originator_topic: str



from __future__ import annotations

import contextlib
from collections import defaultdict
from threading import Event, RLock
from typing import TYPE_CHECKING, Any

from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    InfrastructureFactory,
    IntegrityError,
    ListenNotifySubscription,
    Notification,
    ProcessRecorder,
    StoredEvent,
    Subscription,
    Tracking,
    TrackingRecorder,
)
from eventsourcing.utils import resolve_topic, reversed_keys

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from uuid import UUID


class POPORecorder:
    def __init__(self) -> None:
        self._database_lock = RLock()


class POPOAggregateRecorder(POPORecorder, AggregateRecorder):
    def __init__(self) -> None:
        super().__init__()
        self._stored_events: list[StoredEvent] = []
        self._stored_events_index: dict[str, dict[int, int]] = defaultdict(dict)

    def insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        self._insert_events(stored_events, **kwargs)
        return None

    def _insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        with self._database_lock:
            self._assert_uniqueness(stored_events, **kwargs)
            return self._update_table(stored_events, **kwargs)

    def _assert_uniqueness(
        self, stored_events: Sequence[StoredEvent], **_: Any
    ) -> None:
        new = set()
        for s in stored_events:
            # Check events don't already exist.
            if s.originator_version in self._stored_events_index[str(s.originator_id)]:
                msg = f"Stored event already recorded: {s}"
                raise IntegrityError(msg)
            new.add((s.originator_id, s.originator_version))
        # Check new events are unique.
        if len(new) < len(stored_events):
            msg = f"Stored events are not unique: {stored_events}"
            raise IntegrityError(msg)

    def _update_table(
        self, stored_events: Sequence[StoredEvent], **_: Any
    ) -> Sequence[int] | None:
        notification_ids = []
        for s in stored_events:
            self._stored_events.append(s)
            self._stored_events_index[str(s.originator_id)][s.originator_version] = (
                len(self._stored_events) - 1
            )
            notification_ids.append(len(self._stored_events))
        return notification_ids

    def select_events(
        self,
        originator_id: UUID | str,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Sequence[StoredEvent]:
        with self._database_lock:
            results = []

            index = self._stored_events_index[str(originator_id)]
            positions: Iterable[int]
            positions = reversed_keys(index) if desc else index.keys()
            for p in positions:
                if gt is not None and not p > gt:
                    continue
                if lte is not None and not p <= lte:
                    continue
                s = self._stored_events[index[p]]
                results.append(s)
                if len(results) == limit:
                    break
            return results


class POPOApplicationRecorder(POPOAggregateRecorder, ApplicationRecorder):
    def __init__(self) -> None:
        super().__init__()
        self._listeners: set[Event] = set()

    def insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        notification_ids = self._insert_events(stored_events, **kwargs)
        self._notify_listeners()
        return notification_ids

    def select_notifications(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        with self._database_lock:
            results = []
            if start is None:
                start = 1
                inclusive_of_start = True
            if not inclusive_of_start:
                start += 1
            start = max(start, 1)  # Don't use negative indexes!
            i = start - 1  # Zero-based indexing.
            while True:
                if stop is not None and i > stop - 1:
                    break
                try:
                    s = self._stored_events[i]
                except IndexError:
                    break
                i += 1
                if topics and s.topic not in topics:
                    continue
                n = Notification(
                    id=i,
                    originator_id=s.originator_id,
                    originator_version=s.originator_version,
                    topic=s.topic,
                    state=s.state,
                )
                results.append(n)
                if len(results) == limit:
                    break
            return results

    def max_notification_id(self) -> int | None:
        with self._database_lock:
            return len(self._stored_events) or None

    def subscribe(
        self, gt: int | None = None, topics: Sequence[str] = ()
    ) -> Subscription[ApplicationRecorder]:
        return POPOSubscription(recorder=self, gt=gt, topics=topics)

    def listen(self, event: Event) -> None:
        self._listeners.add(event)

    def unlisten(self, event: Event) -> None:
        with contextlib.suppress(KeyError):
            self._listeners.remove(event)

    def _notify_listeners(self) -> None:
        for listener in self._listeners:
            listener.set()


class POPOSubscription(ListenNotifySubscription[POPOApplicationRecorder]):
    def __init__(
        self,
        recorder: POPOApplicationRecorder,
        gt: int | None = None,
        topics: Sequence[str] = (),
    ) -> None:
        assert isinstance(recorder, POPOApplicationRecorder)
        super().__init__(recorder=recorder, gt=gt, topics=topics)
        self._recorder.listen(self._has_been_notified)

    def stop(self) -> None:
        super().stop()
        self._recorder.unlisten(self._has_been_notified)


class POPOTrackingRecorder(POPORecorder, TrackingRecorder):
    def __init__(self) -> None:
        super().__init__()
        self._max_tracking_ids: dict[str, int | None] = defaultdict(lambda: None)

    def _assert_tracking_uniqueness(self, tracking: Tracking) -> None:
        max_tracking_id = self._max_tracking_ids[tracking.application_name]
        if max_tracking_id is not None and max_tracking_id >= tracking.notification_id:
            msg = (
                f"Tracking notification ID {tracking.notification_id} "
                f"not greater than current max tracking ID {max_tracking_id}"
            )
            raise IntegrityError(msg)

    def insert_tracking(self, tracking: Tracking) -> None:
        with self._database_lock:
            self._assert_tracking_uniqueness(tracking)
            self._insert_tracking(tracking)

    def _insert_tracking(self, tracking: Tracking) -> None:
        self._max_tracking_ids[tracking.application_name] = tracking.notification_id

    def max_tracking_id(self, application_name: str) -> int | None:
        with self._database_lock:
            return self._max_tracking_ids[application_name]


class POPOProcessRecorder(
    POPOTrackingRecorder, POPOApplicationRecorder, ProcessRecorder
):
    def _assert_uniqueness(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> None:
        super()._assert_uniqueness(stored_events, **kwargs)
        t: Tracking | None = kwargs.get("tracking")
        if t:
            self._assert_tracking_uniqueness(t)

    def _update_table(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        notification_ids = super()._update_table(stored_events, **kwargs)
        t: Tracking | None = kwargs.get("tracking")
        if t:
            self._insert_tracking(t)
        return notification_ids


class POPOFactory(InfrastructureFactory[POPOTrackingRecorder]):
    def aggregate_recorder(self, purpose: str = "events") -> AggregateRecorder:
        return POPOAggregateRecorder()

    def application_recorder(self) -> ApplicationRecorder:
        return POPOApplicationRecorder()

    def tracking_recorder(
        self, tracking_recorder_class: type[POPOTrackingRecorder] | None = None
    ) -> POPOTrackingRecorder:
        if tracking_recorder_class is None:
            tracking_recorder_topic = self.env.get(self.TRACKING_RECORDER_TOPIC)
            if tracking_recorder_topic:
                tracking_recorder_class = resolve_topic(tracking_recorder_topic)
            else:
                tracking_recorder_class = POPOTrackingRecorder
        assert tracking_recorder_class is not None
        assert issubclass(tracking_recorder_class, POPOTrackingRecorder)
        return tracking_recorder_class()

    def process_recorder(self) -> ProcessRecorder:
        return POPOProcessRecorder()


Factory = POPOFactory



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast, overload

_T = TypeVar("_T")
_S = TypeVar("_S")

if TYPE_CHECKING:
    from collections.abc import Callable

    class _singledispatchmethod(functools.singledispatchmethod[_T]):  # noqa: N801
        pass

else:

    class _singledispatchmethod(  # noqa: N801
        functools.singledispatchmethod, Generic[_T]
    ):
        pass


class singledispatchmethod(_singledispatchmethod[_T]):  # noqa: N801
    def __init__(self, func: Callable[..., _T]) -> None:
        super().__init__(func)
        self.deferred_registrations: list[
            tuple[type[Any] | Callable[..., _T], Callable[..., _T] | None]
        ] = []

    @overload
    def register(
        self, cls: type[Any], method: None = None
    ) -> Callable[[Callable[..., _T]], Callable[..., _T]]: ...  # pragma: no cover
    @overload
    def register(
        self, cls: Callable[..., _T], method: None = None
    ) -> Callable[..., _T]: ...  # pragma: no cover

    @overload
    def register(
        self, cls: type[Any], method: Callable[..., _T]
    ) -> Callable[..., _T]: ...  # pragma: no cover

    def register(
        self,
        cls: type[Any] | Callable[..., _T],
        method: Callable[..., _T] | None = None,
    ) -> Callable[[Callable[..., _T]], Callable[..., _T]] | Callable[..., _T]:
        """generic_method.register(cls, func) -> func

        Registers a new implementation for the given *cls* on a *generic_method*.
        """
        if isinstance(cls, (classmethod, staticmethod)):
            first_annotation = {}
            for k, v in cls.__func__.__annotations__.items():
                first_annotation[k] = v
                break
            cls.__annotations__ = first_annotation

            # for globals in typing.get_type_hints() in Python 3.8 and 3.9
            if not hasattr(cls, "__wrapped__"):
                cls.__dict__["__wrapped__"] = cls.__func__
                # cls.__wrapped__ = cls.__func__

        try:
            return self.dispatcher.register(cast("type[Any]", cls), func=method)
        except (NameError, TypeError):  # NameError <= Py3.13, TypeError >= Py3.14
            self.deferred_registrations.append(
                (cls, method)  # pyright: ignore [reportArgumentType]
            )
            # TODO: Fix this....
            return method or cls  # pyright: ignore [reportReturnType]

    def __get__(self, obj: _S, cls: type[_S] | None = None) -> Callable[..., _T]:
        for registered_cls, registered_method in self.deferred_registrations:
            self.dispatcher.register(
                cast("type[Any]", registered_cls), func=registered_method
            )
        self.deferred_registrations = []
        return super().__get__(obj, cls=cls)



from __future__ import annotations

import json
import queue
import sys
import typing
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Callable, Hashable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from functools import lru_cache
from threading import Condition, Event, Lock, Semaphore, Thread, Timer
from time import monotonic, sleep, time
from types import GenericAlias, ModuleType, TracebackType
from typing import Any, Generic, cast
from uuid import UUID

from typing_extensions import Self, TypeVar

from eventsourcing.domain import (
    DomainEventProtocol,
    EventSourcingError,
    HasOriginatorIDVersion,
    TAggregateID,
)
from eventsourcing.utils import (
    Environment,
    EnvType,
    TopicError,
    get_topic,
    resolve_topic,
    strtobool,
)

# Backport Queue shutdown feature - remove when dropping support for Python 3.12.
_T = TypeVar("_T")
if sys.version_info[0:2] < (3, 13):  # pragma: no cover

    class ShutDown(Exception):  # noqa: N818  # pyright: ignore[reportRedeclaration]
        """Raised when put/get with shut-down queue."""

    class Queue(queue.Queue[_T]):  # pyright: ignore[reportRedeclaration]
        def __init__(self, maxsize: int = 0):
            super().__init__(maxsize)
            # Queue shutdown state
            self.is_shutdown = False

        def put(
            self,
            item: _T,
            block: bool = True,  # noqa: FBT001,FBT002
            timeout: float | None = None,
        ) -> None:
            """Put an item into the queue.

            If optional args 'block' is true and 'timeout' is None (the default),
            block if necessary until a free slot is available. If 'timeout' is
            a non-negative number, it blocks at most 'timeout' seconds and raises
            the Full exception if no free slot was available within that time.
            Otherwise ('block' is false), put an item on the queue if a free slot
            is immediately available, else raise the Full exception ('timeout'
            is ignored in that case).

            Raises ShutDown if the queue has been shut down.
            """
            with self.not_full:
                if self.is_shutdown:
                    raise ShutDown
                if self.maxsize > 0:
                    if not block:
                        if self._qsize() >= self.maxsize:
                            raise queue.Full
                    elif timeout is None:
                        while self._qsize() >= self.maxsize:
                            self.not_full.wait()
                            if self.is_shutdown:
                                raise ShutDown
                    elif timeout < 0:
                        msg = "'timeout' must be a non-negative number"
                        raise ValueError(msg)
                    else:
                        endtime = time() + timeout
                        while self._qsize() >= self.maxsize:
                            remaining = endtime - time()
                            if remaining <= 0.0:
                                raise queue.Full
                            self.not_full.wait(remaining)
                            if self.is_shutdown:
                                raise ShutDown
                self._put(item)
                self.unfinished_tasks += 1
                self.not_empty.notify()

        def get(
            self,
            block: bool = True,  # noqa: FBT001,FBT002
            timeout: float | None = None,
        ) -> _T:
            """Remove and return an item from the queue.

            If optional args 'block' is true and 'timeout' is None (the default),
            block if necessary until an item is available. If 'timeout' is
            a non-negative number, it blocks at most 'timeout' seconds and raises
            the Empty exception if no item was available within that time.
            Otherwise ('block' is false), return an item if one is immediately
            available, else raise the Empty exception ('timeout' is ignored
            in that case).

            Raises ShutDown if the queue has been shut down and is empty,
            or if the queue has been shut down immediately.
            """
            with self.not_empty:
                if self.is_shutdown and not self._qsize():
                    raise ShutDown
                if not block:
                    if not self._qsize():
                        raise queue.Empty
                elif timeout is None:
                    while not self._qsize():
                        self.not_empty.wait()
                        if self.is_shutdown and not self._qsize():
                            raise ShutDown
                elif timeout < 0:
                    msg = "'timeout' must be a non-negative number"
                    raise ValueError(msg)
                else:
                    endtime = time() + timeout
                    while not self._qsize():
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise queue.Empty
                        self.not_empty.wait(remaining)
                        if self.is_shutdown and not self._qsize():
                            raise ShutDown
                item = self._get()
                self.not_full.notify()
                return item

        def shutdown(
            self,
            immediate: bool = False,  # noqa: FBT001,FBT002
        ) -> None:
            """Shut-down the queue, making queue gets and puts raise ShutDown.

            By default, gets will only raise once the queue is empty. Set
            'immediate' to True to make gets raise immediately instead.

            All blocked callers of put() and get() will be unblocked. If
            'immediate', a task is marked as done for each item remaining in
            the queue, which may unblock callers of join().
            """
            with self.mutex:
                self.is_shutdown = True
                if immediate:
                    while self._qsize():
                        self._get()
                        if self.unfinished_tasks > 0:
                            self.unfinished_tasks -= 1
                    # release all blocked threads in `join()`
                    self.all_tasks_done.notify_all()
                # All getters need to re-check queue-empty to raise ShutDown
                self.not_empty.notify_all()
                self.not_full.notify_all()

else:  # pragma: no cover
    Queue = queue.Queue  # pyright: ignore[reportAssignmentType]
    ShutDown = queue.ShutDown  # pyright: ignore[reportAttributeAccessIssue]


class Transcoding(ABC):
    """Abstract base class for custom transcodings."""

    type: type
    name: str

    @abstractmethod
    def encode(self, obj: Any) -> Any:
        """Encodes given object."""

    @abstractmethod
    def decode(self, data: Any) -> Any:
        """Decodes encoded object."""


class Transcoder(ABC):
    """Abstract base class for transcoders."""

    @abstractmethod
    def encode(self, obj: Any) -> bytes:
        """Encodes obj as bytes."""

    @abstractmethod
    def decode(self, data: bytes) -> Any:
        """Decodes obj from bytes."""


class TranscodingNotRegisteredError(EventSourcingError, TypeError):
    """Raised when a transcoding isn't registered with JSONTranscoder."""


class JSONTranscoder(Transcoder):
    """Extensible transcoder that uses the Python :mod:`json` module."""

    def __init__(self) -> None:
        self.types: dict[type, Transcoding] = {}
        self.names: dict[str, Transcoding] = {}
        self.encoder = json.JSONEncoder(
            default=self._encode_obj,
            separators=(",", ":"),
            ensure_ascii=False,
        )
        self.decoder = json.JSONDecoder(object_hook=self._decode_obj)

    def register(self, transcoding: Transcoding) -> None:
        """Registers given transcoding with the transcoder."""
        self.types[transcoding.type] = transcoding
        self.names[transcoding.name] = transcoding

    def encode(self, obj: Any) -> bytes:
        """Encodes given object as a bytes array."""
        return self.encoder.encode(obj).encode("utf8")

    def decode(self, data: bytes) -> Any:
        """Decodes bytes array as previously encoded object."""
        return self.decoder.decode(data.decode("utf8"))

    def _encode_obj(self, o: Any) -> dict[str, Any]:
        try:
            transcoding = self.types[type(o)]
        except KeyError:
            msg = (
                f"Object of type {type(o)} is not "
                "serializable. Please define and register "
                "a custom transcoding for this type."
            )
            raise TranscodingNotRegisteredError(msg) from None
        else:
            return {
                "_type_": transcoding.name,
                "_data_": transcoding.encode(o),
            }

    def _decode_obj(self, d: dict[str, Any]) -> Any:
        if len(d) == 2:
            try:
                _type_ = d["_type_"]
            except KeyError:
                return d
            else:
                try:
                    _data_ = d["_data_"]
                except KeyError:
                    return d
                else:
                    try:
                        transcoding = self.names[cast("str", _type_)]
                    except KeyError as e:
                        msg = (
                            f"Data serialized with name '{cast('str', _type_)}' is not "
                            "deserializable. Please register a "
                            "custom transcoding for this type."
                        )
                        raise TranscodingNotRegisteredError(msg) from e
                    else:
                        return transcoding.decode(_data_)
        else:
            return d


class UUIDAsHex(Transcoding):
    """Transcoding that represents :class:`UUID` objects as hex values."""

    type = UUID
    name = "uuid_hex"

    def encode(self, obj: UUID) -> str:
        return obj.hex

    def decode(self, data: str) -> UUID:
        assert isinstance(data, str)
        return UUID(data)


class DecimalAsStr(Transcoding):
    """Transcoding that represents :class:`Decimal` objects as strings."""

    type = Decimal
    name = "decimal_str"

    def encode(self, obj: Decimal) -> str:
        return str(obj)

    def decode(self, data: str) -> Decimal:
        return Decimal(data)


class DatetimeAsISO(Transcoding):
    """Transcoding that represents :class:`datetime` objects as ISO strings."""

    type = datetime
    name = "datetime_iso"

    def encode(self, obj: datetime) -> str:
        return obj.isoformat()

    def decode(self, data: str) -> datetime:
        assert isinstance(data, str)
        return datetime.fromisoformat(data)


@dataclass(frozen=True)
class StoredEvent:
    """Frozen dataclass that represents :class:`~eventsourcing.domain.DomainEvent`
    objects, such as aggregate :class:`~eventsourcing.domain.Aggregate.Event`
    objects and :class:`~eventsourcing.domain.Snapshot` objects.
    """

    originator_id: UUID | str
    """ID of the originating aggregate."""
    originator_version: int
    """Position in an aggregate sequence."""
    topic: str
    """Topic of a domain event object class."""
    state: bytes
    """Serialised state of a domain event object."""


class Compressor(ABC):
    """Base class for compressors."""

    @abstractmethod
    def compress(self, data: bytes) -> bytes:
        """Compress bytes."""

    @abstractmethod
    def decompress(self, data: bytes) -> bytes:
        """Decompress bytes."""


class Cipher(ABC):
    """Base class for ciphers."""

    @abstractmethod
    def __init__(self, environment: Environment):
        """Initialises cipher with given environment."""

    @abstractmethod
    def encrypt(self, plaintext: bytes) -> bytes:
        """Return ciphertext for given plaintext."""

    @abstractmethod
    def decrypt(self, ciphertext: bytes) -> bytes:
        """Return plaintext for given ciphertext."""


class MapperDeserialisationError(EventSourcingError, ValueError):
    """Raised when deserialization fails in a Mapper."""


TAggregateIDType = TypeVar("TAggregateIDType", type[UUID], type[str])


class Mapper(Generic[TAggregateID]):
    """Converts between domain event objects and :class:`StoredEvent` objects.

    Uses a :class:`Transcoder`, and optionally a cryptographic cipher and compressor.
    """

    def __init__(
        self,
        transcoder: Transcoder,
        compressor: Compressor | None = None,
        cipher: Cipher | None = None,
    ):
        self.transcoder = transcoder
        self.compressor = compressor
        self.cipher = cipher

    def to_stored_event(
        self, domain_event: DomainEventProtocol[TAggregateID]
    ) -> StoredEvent:
        """Converts the given domain event to a :class:`StoredEvent` object."""
        topic = get_topic(domain_event.__class__)
        event_state = dict(vars(domain_event))
        originator_id = event_state.pop("originator_id")
        originator_version = event_state.pop("originator_version")
        class_version = getattr(type(domain_event), "class_version", 1)
        if class_version > 1:
            event_state["class_version"] = class_version
        stored_state = self.transcoder.encode(event_state)
        if self.compressor:
            stored_state = self.compressor.compress(stored_state)
        if self.cipher:
            stored_state = self.cipher.encrypt(stored_state)
        return StoredEvent(
            originator_id=originator_id,
            originator_version=originator_version,
            topic=topic,
            state=stored_state,
        )

    def to_domain_event(
        self, stored_event: StoredEvent
    ) -> DomainEventProtocol[TAggregateID]:
        """Converts the given :class:`StoredEvent` to a domain event object."""
        cls = resolve_topic(stored_event.topic)

        stored_state = stored_event.state
        try:
            if self.cipher:
                stored_state = self.cipher.decrypt(stored_state)
            if self.compressor:
                stored_state = self.compressor.decompress(stored_state)
            event_state: dict[str, Any] = self.transcoder.decode(stored_state)
        except Exception as e:
            msg = (
                f"Failed to deserialise state of stored event with "
                f"topic '{stored_event.topic}', "
                f"originator_id '{stored_event.originator_id}' and "
                f"originator_version {stored_event.originator_version}: {e}"
            )
            raise MapperDeserialisationError(msg) from e

        id_convertor = find_id_convertor(
            cls, cast(Hashable, type(stored_event.originator_id))
        )
        # print("ID of convertor:", id(convertor))
        event_state["originator_id"] = id_convertor(stored_event.originator_id)
        event_state["originator_version"] = stored_event.originator_version
        class_version = getattr(cls, "class_version", 1)
        from_version = event_state.pop("class_version", 1)
        while from_version < class_version:
            getattr(cls, f"upcast_v{from_version}_v{from_version + 1}")(event_state)
            from_version += 1

        domain_event = object.__new__(cls)
        object.__setattr__(domain_event, "__dict__", event_state)
        return domain_event


@lru_cache
def find_id_convertor(
    domain_event_cls: type[object], originator_id_cls: type[UUID | str]
) -> Callable[[UUID | str], UUID | str]:
    # Try to find the originator_id type.
    if issubclass(domain_event_cls, HasOriginatorIDVersion):
        # For classes that inherit CanMutateAggregate, and don't use a different
        # mapper, then assume they aren't overriding __init_subclass__ is a way
        # that prevents 'originator_id_type' being found from type arguments and
        # set on the class.
        # TODO: Write a test where a custom class does override __init_subclass__
        #  so that the next line will cause an AssertionError. Then fix this code.
        if domain_event_cls.originator_id_type is None:
            msg = "originator_id_type cannot be None"
            raise TypeError(msg)
        originator_id_type = domain_event_cls.originator_id_type
    else:
        # Otherwise look for annotations.
        for cls in domain_event_cls.__mro__:
            try:
                annotation = cls.__annotations__["originator_id"]
            except (KeyError, AttributeError):  # noqa: PERF203
                continue
            else:
                valid_annotations = {
                    str: str,
                    UUID: UUID,
                    "str": str,
                    "UUID": UUID,
                    "uuid.UUID": UUID,
                }
                if annotation not in valid_annotations:
                    msg = f"originator_id annotation on {cls} is not either UUID or str"
                    raise TypeError(msg)
                assert annotation in valid_annotations, annotation
                originator_id_type = valid_annotations[annotation]
                break
        else:
            msg = (
                f"Neither event class {domain_event_cls}"
                f"nor its bases have an originator_id annotation"
            )
            raise TypeError(msg)

    if originator_id_cls is str and originator_id_type is UUID:
        convertor = str_to_uuid_convertor
    else:
        convertor = pass_through_convertor
    return convertor


def str_to_uuid_convertor(originator_id: UUID | str) -> UUID | str:
    assert isinstance(originator_id, str)
    return UUID(originator_id)


def pass_through_convertor(originator_id: UUID | str) -> UUID | str:
    return originator_id


class RecordConflictError(EventSourcingError):
    """Legacy exception, replaced with IntegrityError."""


class PersistenceError(EventSourcingError):
    """The base class of the other exceptions in this module.

    Exception class names follow https://www.python.org/dev/peps/pep-0249/#exceptions
    """


class InterfaceError(PersistenceError):
    """Exception raised for errors that are related to the database
    interface rather than the database itself.
    """


class DatabaseError(PersistenceError):
    """Exception raised for errors that are related to the database."""


class DataError(DatabaseError):
    """Exception raised for errors that are due to problems with the
    processed data like division by zero, numeric value out of range, etc.
    """


class OperationalError(DatabaseError):
    """Exception raised for errors that are related to the database's
    operation and not necessarily under the control of the programmer,
    e.g. an unexpected disconnect occurs, the data source name is not
    found, a transaction could not be processed, a memory allocation
    error occurred during processing, etc.
    """


class IntegrityError(DatabaseError, RecordConflictError):
    """Exception raised when the relational integrity of the
    database is affected, e.g. a foreign key check fails.
    """


class InternalError(DatabaseError):
    """Exception raised when the database encounters an internal
    error, e.g. the cursor is not valid anymore, the transaction
    is out of sync, etc.
    """


class ProgrammingError(DatabaseError):
    """Exception raised for database programming errors, e.g. table
    not found or already exists, syntax error in the SQL statement,
    wrong number of parameters specified, etc.
    """


class NotSupportedError(DatabaseError):
    """Exception raised in case a method or database API was used
    which is not supported by the database, e.g. calling the
    rollback() method on a connection that does not support
    transaction or has transactions turned off.
    """


class WaitInterruptedError(PersistenceError):
    """Raised when waiting for a tracking record is interrupted."""


class Recorder:
    pass


class AggregateRecorder(Recorder, ABC):
    """Abstract base class for inserting and selecting stored events."""

    @abstractmethod
    def insert_events(
        self, stored_events: Sequence[StoredEvent], **kwargs: Any
    ) -> Sequence[int] | None:
        """Writes stored events into database."""

    @abstractmethod
    def select_events(
        self,
        originator_id: UUID | str,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Sequence[StoredEvent]:
        """Reads stored events from database."""


@dataclass(frozen=True)
class Notification(StoredEvent):
    """Frozen dataclass that represents domain event notifications."""

    id: int
    """Position in an application sequence."""


class ApplicationRecorder(AggregateRecorder):
    """Abstract base class for recording events in both aggregate
    and application sequences.
    """

    @abstractmethod
    def select_notifications(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        """Returns a list of Notification objects representing events from an
        application sequence. If `inclusive_of_start` is True (the default),
        the returned Notification objects will have IDs greater than or equal
        to `start` and less than or equal to `stop`. If `inclusive_of_start`
        is False, the Notification objects will have IDs greater than `start`
        and less than or equal to `stop`.
        """

    @abstractmethod
    def max_notification_id(self) -> int | None:
        """Returns the largest notification ID in an application sequence,
        or None if no stored events have been recorded.
        """

    @abstractmethod
    def subscribe(
        self, gt: int | None = None, topics: Sequence[str] = ()
    ) -> Subscription[ApplicationRecorder]:
        """Returns an iterator of Notification objects representing events from an
        application sequence.

        The iterator will block after the last recorded event has been yielded, but
        will then continue yielding newly recorded events when they are recorded.

        Notifications will have IDs greater than the optional `gt` argument.
        """


class TrackingRecorder(Recorder, ABC):
    """Abstract base class for recorders that record tracking
    objects atomically with other state.
    """

    @abstractmethod
    def insert_tracking(self, tracking: Tracking) -> None:
        """Records a tracking object."""

    @abstractmethod
    def max_tracking_id(self, application_name: str) -> int | None:
        """Returns the largest notification ID across all recorded tracking objects
        for the named application, or None if no tracking objects have been recorded.
        """

    def has_tracking_id(
        self, application_name: str, notification_id: int | None
    ) -> bool:
        """Returns True if given notification_id is None or a tracking
        object with the given application_name and a notification ID greater
        than or equal to the given notification_id has been recorded.
        """
        if notification_id is None:
            return True
        max_tracking_id = self.max_tracking_id(application_name)
        return max_tracking_id is not None and max_tracking_id >= notification_id

    def wait(
        self,
        application_name: str,
        notification_id: int | None,
        timeout: float = 1.0,
        interrupt: Event | None = None,
    ) -> None:
        """Block until a tracking object with the given application name and a
        notification ID greater than equal to the given value has been recorded.

        Polls max_tracking_id() with exponential backoff until the timeout
        is reached, or until the optional interrupt event is set.

        The timeout argument should be a floating point number specifying a
        timeout for the operation in seconds (or fractions thereof). The default
        is 1.0 seconds.

        Raises TimeoutError if the timeout is reached.

        Raises WaitInterruptError if the `interrupt` is set before `timeout` is reached.
        """
        deadline = monotonic() + timeout
        sleep_interval_ms = 100.0
        max_sleep_interval_ms = 800.0
        while True:
            if self.has_tracking_id(application_name, notification_id):
                break
            if interrupt:
                if interrupt.wait(timeout=sleep_interval_ms / 1000):
                    raise WaitInterruptedError
            else:
                sleep(sleep_interval_ms / 1000)
            remaining = deadline - monotonic()
            if remaining < 0:
                msg = (
                    f"Timed out waiting for notification {notification_id} "
                    f"from application '{application_name}' to be processed"
                )
                raise TimeoutError(msg)
            sleep_interval_ms = min(
                sleep_interval_ms * 2, remaining * 1000, max_sleep_interval_ms
            )


class ProcessRecorder(TrackingRecorder, ApplicationRecorder, ABC):
    pass


@dataclass(frozen=True)
class Recording(Generic[TAggregateID]):
    """Represents the recording of a domain event."""

    domain_event: DomainEventProtocol[TAggregateID]
    """The domain event that has been recorded."""
    notification: Notification
    """A Notification that represents the domain event in the application sequence."""


class EventStore(Generic[TAggregateID]):
    """Stores and retrieves domain events."""

    def __init__(
        self,
        mapper: Mapper[TAggregateID],
        recorder: AggregateRecorder,
    ):
        self.mapper: Mapper[TAggregateID] = mapper
        self.recorder = recorder

    def put(
        self, domain_events: Sequence[DomainEventProtocol[TAggregateID]], **kwargs: Any
    ) -> list[Recording[TAggregateID]]:
        """Stores domain events in aggregate sequence."""
        stored_events = list(map(self.mapper.to_stored_event, domain_events))
        recordings = []
        notification_ids = self.recorder.insert_events(stored_events, **kwargs)
        if notification_ids:
            assert len(notification_ids) == len(stored_events)
            for d, s, n_id in zip(
                domain_events, stored_events, notification_ids, strict=True
            ):
                recordings.append(
                    Recording(
                        d,
                        Notification(
                            originator_id=s.originator_id,
                            originator_version=s.originator_version,
                            topic=s.topic,
                            state=s.state,
                            id=n_id,
                        ),
                    )
                )
        return recordings

    def get(
        self,
        originator_id: TAggregateID,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Iterator[DomainEventProtocol[TAggregateID]]:
        """Retrieves domain events from aggregate sequence."""
        return map(
            self.mapper.to_domain_event,
            self.recorder.select_events(
                originator_id=originator_id,
                gt=gt,
                lte=lte,
                desc=desc,
                limit=limit,
            ),
        )


TTrackingRecorder = TypeVar(
    "TTrackingRecorder", bound=TrackingRecorder, default=TrackingRecorder
)


class InfrastructureFactoryError(EventSourcingError):
    """Raised when an infrastructure factory cannot be created."""


class BaseInfrastructureFactory(ABC, Generic[TTrackingRecorder]):
    """Abstract base class for infrastructure factories."""

    PERSISTENCE_MODULE = "PERSISTENCE_MODULE"
    TRANSCODER_TOPIC = "TRANSCODER_TOPIC"
    CIPHER_TOPIC = "CIPHER_TOPIC"
    COMPRESSOR_TOPIC = "COMPRESSOR_TOPIC"

    def __init__(self, env: Environment | EnvType | None):
        """Initialises infrastructure factory object with given application name."""
        self.env = env if isinstance(env, Environment) else Environment(env=env)
        self._is_entered = False

    def __enter__(self) -> Self:
        self._is_entered = True
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._is_entered = False

    def close(self) -> None:
        """Closes any database connections, and anything else that needs closing."""

    @classmethod
    def construct(
        cls: type[Self],
        env: Environment | None = None,
    ) -> Self:
        """Constructs concrete infrastructure factory for given
        named application. Reads and resolves persistence
        topic from environment variable 'PERSISTENCE_MODULE'.
        """
        factory_cls: type[Self]
        if env is None:
            env = Environment()
        topic = (
            env.get(
                cls.PERSISTENCE_MODULE,
                "",
            )
            or env.get(
                "INFRASTRUCTURE_FACTORY",  # Legacy.
                "",
            )
            or env.get(
                "FACTORY_TOPIC",  # Legacy.
                "",
            )
            or "eventsourcing.popo"
        )
        try:
            obj: type[Self] | ModuleType = resolve_topic(topic)
        except TopicError as e:
            msg = (
                "Failed to resolve persistence module topic: "
                f"'{topic}' from environment "
                f"variable '{cls.PERSISTENCE_MODULE}'"
            )
            raise InfrastructureFactoryError(msg) from e

        if isinstance(obj, ModuleType):
            # Find the factory in the module.
            factory_classes = set[type[Self]]()
            for member in obj.__dict__.values():
                # Look for classes...
                if not isinstance(member, type):
                    continue
                # Issue with Python 3.9 and 3.10.
                if isinstance(member, GenericAlias):
                    continue  # pragma: no cover (for Python > 3.10 only)
                if not issubclass(member, cls):
                    continue
                if getattr(member, "__parameters__", None):
                    continue
                factory_classes.add(member)

            if len(factory_classes) == 1:
                factory_cls = next(iter(factory_classes))
            else:
                msg = (
                    f"Found {len(factory_classes)} infrastructure factory classes in"
                    f" '{topic}', expected 1."
                )
                raise InfrastructureFactoryError(msg)
        elif isinstance(obj, type) and issubclass(obj, cls):
            factory_cls = obj
        else:
            msg = (
                f"Topic '{topic}' didn't resolve to a persistence module "
                f"or infrastructure factory class: {obj}"
            )
            raise InfrastructureFactoryError(msg)
        return factory_cls(env=env)

    def transcoder(
        self,
    ) -> Transcoder:
        """Constructs a transcoder."""
        transcoder_topic = self.env.get(self.TRANSCODER_TOPIC)
        if transcoder_topic:
            transcoder_class: type[Transcoder] = resolve_topic(transcoder_topic)
        else:
            transcoder_class = JSONTranscoder
        return transcoder_class()

    def cipher(self) -> Cipher | None:
        """Reads environment variables 'CIPHER_TOPIC'
        and 'CIPHER_KEY' to decide whether or not
        to construct a cipher.
        """
        cipher_topic = self.env.get(self.CIPHER_TOPIC)
        cipher: Cipher | None = None
        default_cipher_topic = "eventsourcing.cipher:AESCipher"
        if self.env.get("CIPHER_KEY") and not cipher_topic:
            cipher_topic = default_cipher_topic

        if cipher_topic:
            cipher_cls: type[Cipher] = resolve_topic(cipher_topic)
            cipher = cipher_cls(self.env)

        return cipher

    def compressor(self) -> Compressor | None:
        """Reads environment variable 'COMPRESSOR_TOPIC' to
        decide whether or not to construct a compressor.
        """
        compressor: Compressor | None = None
        compressor_topic = self.env.get(self.COMPRESSOR_TOPIC)
        if compressor_topic:
            compressor_cls: type[Compressor] | Compressor = resolve_topic(
                compressor_topic
            )
            if isinstance(compressor_cls, type):
                compressor = compressor_cls()
            else:
                compressor = compressor_cls
        return compressor


class InfrastructureFactory(BaseInfrastructureFactory[TTrackingRecorder]):
    """Abstract base class for Application factories."""

    MAPPER_TOPIC = "MAPPER_TOPIC"
    IS_SNAPSHOTTING_ENABLED = "IS_SNAPSHOTTING_ENABLED"
    APPLICATION_RECORDER_TOPIC = "APPLICATION_RECORDER_TOPIC"
    TRACKING_RECORDER_TOPIC = "TRACKING_RECORDER_TOPIC"
    PROCESS_RECORDER_TOPIC = "PROCESS_RECORDER_TOPIC"

    def mapper(
        self,
        transcoder: Transcoder | None = None,
        mapper_class: type[Mapper[TAggregateID]] | None = None,
    ) -> Mapper[TAggregateID]:
        """Constructs a mapper."""
        # Resolve MAPPER_TOPIC if no given class.
        if mapper_class is None:
            mapper_topic = self.env.get(self.MAPPER_TOPIC)
            mapper_class = (
                resolve_topic(mapper_topic) if mapper_topic else Mapper[TAggregateID]
            )

        # Check we have a mapper class.
        assert mapper_class is not None
        origin_mapper_class = typing.get_origin(mapper_class) or mapper_class
        assert isinstance(origin_mapper_class, type), mapper_class
        assert issubclass(origin_mapper_class, Mapper), mapper_class

        # Construct and return a mapper.
        return mapper_class(
            transcoder=transcoder or self.transcoder(),
            cipher=self.cipher(),
            compressor=self.compressor(),
        )

    def event_store(
        self,
        mapper: Mapper[TAggregateID] | None = None,
        recorder: AggregateRecorder | None = None,
    ) -> EventStore[TAggregateID]:
        """Constructs an event store."""
        return EventStore(
            mapper=mapper or self.mapper(),
            recorder=recorder or self.application_recorder(),
        )

    @abstractmethod
    def aggregate_recorder(self, purpose: str = "events") -> AggregateRecorder:
        """Constructs an aggregate recorder."""

    @abstractmethod
    def application_recorder(self) -> ApplicationRecorder:
        """Constructs an application recorder."""

    @abstractmethod
    def tracking_recorder(
        self, tracking_recorder_class: type[TTrackingRecorder] | None = None
    ) -> TTrackingRecorder:
        """Constructs a tracking recorder."""

    @abstractmethod
    def process_recorder(self) -> ProcessRecorder:
        """Constructs a process recorder."""

    def is_snapshotting_enabled(self) -> bool:
        """Decides whether or not snapshotting is enabled by
        reading environment variable 'IS_SNAPSHOTTING_ENABLED'.
        Snapshotting is not enabled by default.
        """
        return strtobool(self.env.get(self.IS_SNAPSHOTTING_ENABLED, "no"))


@dataclass(frozen=True)
class Tracking:
    """Frozen dataclass representing the position of a domain
    event :class:`Notification` in an application's notification log.
    """

    application_name: str
    notification_id: int


Params = Sequence[Any] | Mapping[str, Any]


class Cursor(ABC):
    @abstractmethod
    def execute(self, statement: str | bytes, params: Params | None = None) -> None:
        """Executes given statement."""

    @abstractmethod
    def fetchall(self) -> Any:
        """Fetches all results."""

    @abstractmethod
    def fetchone(self) -> Any:
        """Fetches one result."""


TCursor = TypeVar("TCursor", bound=Cursor)


class Connection(ABC, Generic[TCursor]):
    def __init__(self, max_age: float | None = None) -> None:
        self._closed = False
        self._closing = Event()
        self._close_lock = Lock()
        self.in_use = Lock()
        self.in_use.acquire()
        if max_age is not None:
            self._max_age_timer: Timer | None = Timer(
                interval=max_age,
                function=self._close_when_not_in_use,
            )
            self._max_age_timer.daemon = True
            self._max_age_timer.start()
        else:
            self._max_age_timer = None
        self.is_writer: bool | None = None

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def closing(self) -> bool:
        return self._closing.is_set()

    @abstractmethod
    def commit(self) -> None:
        """Commits transaction."""

    @abstractmethod
    def rollback(self) -> None:
        """Rolls back transaction."""

    @abstractmethod
    def cursor(self) -> TCursor:
        """Creates new cursor."""

    def close(self) -> None:
        with self._close_lock:
            self._close()

    @abstractmethod
    def _close(self) -> None:
        self._closed = True
        if self._max_age_timer:
            self._max_age_timer.cancel()

    def _close_when_not_in_use(self) -> None:
        self._closing.set()
        with self.in_use:
            if not self._closed:
                self.close()


TConnection = TypeVar("TConnection", bound=Connection[Any])


class ConnectionPoolClosedError(EventSourcingError):
    """Raised when using a connection pool that is already closed."""


class ConnectionNotFromPoolError(EventSourcingError):
    """Raised when putting a connection in the wrong pool."""


class ConnectionUnavailableError(OperationalError, TimeoutError):
    """Raised when a request to get a connection from a
    connection pool times out.
    """


class ConnectionPool(ABC, Generic[TConnection]):
    def __init__(
        self,
        *,
        pool_size: int = 5,
        max_overflow: int = 10,
        pool_timeout: float = 30.0,
        max_age: float | None = None,
        pre_ping: bool = False,
        mutually_exclusive_read_write: bool = False,
    ) -> None:
        """Initialises a new connection pool.

        The 'pool_size' argument specifies the maximum number of connections
        that will be put into the pool when connections are returned. The
        default value is 5

        The 'max_overflow' argument specifies the additional number of
        connections that can be issued by the pool, above the 'pool_size'.
        The default value is 10.

        The 'pool_timeout' argument specifies the maximum time in seconds
        to keep requests for connections waiting. Connections are kept
        waiting if the number of connections currently in use is not less
        than the sum of 'pool_size' and 'max_overflow'. The default value
        is 30.0

        The 'max_age' argument specifies the time in seconds until a
        connection will automatically be closed. Connections are only closed
        in this way after are not in use. Connections that are in use will
        not be closed automatically. The default value in None, meaning
        connections will not be automatically closed in this way.

        The 'mutually_exclusive_read_write' argument specifies whether
        requests for connections for writing whilst connections for reading
        are in use. It also specifies whether requests for connections for reading
        will be kept waiting whilst a connection for writing is in use. The default
        value is false, meaning reading and writing will not be mutually exclusive
        in this way.
        """
        self.pool_size = pool_size
        self.max_overflow = max_overflow
        self.pool_timeout = pool_timeout
        self.max_age = max_age
        self.pre_ping = pre_ping
        self._pool: deque[TConnection] = deque()
        self._in_use: dict[int, TConnection] = {}
        self._get_semaphore = Semaphore()
        self._put_condition = Condition()
        self._no_readers = Condition()
        self._num_readers: int = 0
        self._writer_lock = Lock()
        self._num_writers: int = 0
        self._mutually_exclusive_read_write = mutually_exclusive_read_write
        self._closed = False

    @property
    def closed(self) -> bool:
        return self._closed

    @property
    def num_in_use(self) -> int:
        """Indicates the total number of connections currently in use."""
        with self._put_condition:
            return self._num_in_use

    @property
    def _num_in_use(self) -> int:
        return len(self._in_use)

    @property
    def num_in_pool(self) -> int:
        """Indicates the number of connections currently in the pool."""
        with self._put_condition:
            return self._num_in_pool

    @property
    def _num_in_pool(self) -> int:
        return len(self._pool)

    @property
    def _is_pool_full(self) -> bool:
        return self._num_in_pool >= self.pool_size

    @property
    def _is_use_full(self) -> bool:
        return self._num_in_use >= self.pool_size + self.max_overflow

    def get_connection(
        self, timeout: float | None = None, *, is_writer: bool | None = None
    ) -> TConnection:
        """Issues connections, or raises ConnectionPoolExhausted error.
        Provides "fairness" on attempts to get connections, meaning that
        connections are issued in the same order as they are requested.

        The 'timeout' argument overrides the timeout specified
        by the constructor argument 'pool_timeout'. The default
        value is None, meaning the 'pool_timeout' argument will
        not be overridden.

        The optional 'is_writer' argument can be used to request
        a connection for writing (true), and request a connection
        for reading (false). If the value of this argument is None,
        which is the default, the writing and reading interlocking
        mechanism is not activated. Only one connection for writing
        will be issued, which means requests for connections for
        writing are kept waiting whilst another connection for writing
        is in use.

        If reading and writing are mutually exclusive, requsts for
        connections for writing are kept waiting whilst connections
        for reading are in use, and requests for connections for reading
        are kept waiting whilst a connection for writing is in use.
        """
        # Make sure we aren't dealing with a closed pool.
        if self._closed:
            raise ConnectionPoolClosedError

        # Decide the timeout for getting a connection.
        timeout = self.pool_timeout if timeout is None else timeout

        # Remember when we started trying to get a connection.
        started = time()

        # Join queue of threads waiting to get a connection ("fairness").
        if self._get_semaphore.acquire(timeout=timeout):
            try:
                # If connection is for writing, get write lock and wait for no readers.
                if is_writer is True:
                    if not self._writer_lock.acquire(
                        timeout=self._time_remaining(timeout, started)
                    ):
                        msg = "Timed out waiting for return of writer"
                        raise ConnectionUnavailableError(msg)
                    if self._mutually_exclusive_read_write:
                        with self._no_readers:
                            if self._num_readers > 0 and not self._no_readers.wait(
                                timeout=self._time_remaining(timeout, started)
                            ):
                                self._writer_lock.release()
                                msg = "Timed out waiting for return of reader"
                                raise ConnectionUnavailableError(msg)
                    self._num_writers += 1

                # If connection is for reading, and writing excludes reading,
                # then wait for the writer lock, and increment number of readers.
                elif is_writer is False:
                    if self._mutually_exclusive_read_write:
                        if not self._writer_lock.acquire(
                            timeout=self._time_remaining(timeout, started)
                        ):
                            msg = "Timed out waiting for return of writer"
                            raise ConnectionUnavailableError(msg)
                        self._writer_lock.release()
                    with self._no_readers:
                        self._num_readers += 1

                # Actually try to get a connection withing the time remaining.
                conn = self._get_connection(
                    timeout=self._time_remaining(timeout, started)
                )

                # Remember if this connection is for reading or writing.
                conn.is_writer = is_writer

                # Return the connection.
                return conn
            finally:
                self._get_semaphore.release()
        else:
            # Timed out waiting for semaphore.
            msg = "Timed out waiting for connection pool semaphore"
            raise ConnectionUnavailableError(msg)

    def _get_connection(self, timeout: float = 0.0) -> TConnection:
        """Gets or creates connections from pool within given
        time, otherwise raises a "pool exhausted" error.

        Waits for connections to be returned if the pool
        is fully used. And optionally ensures a connection
        is usable before returning a connection for use.

        Tracks use of connections, and number of readers.
        """
        started = time()
        # Get lock on tracking usage of connections.
        with self._put_condition:
            # Try to get a connection from the pool.
            try:
                conn = self._pool.popleft()
            except IndexError:
                # Pool is empty, but are connections fully used?
                if self._is_use_full:
                    # Fully used, so wait for a connection to be returned.
                    if self._put_condition.wait(
                        timeout=self._time_remaining(timeout, started)
                    ):
                        # Connection has been returned, so try again.
                        return self._get_connection(
                            timeout=self._time_remaining(timeout, started)
                        )
                    # Timed out waiting for a connection to be returned.
                    msg = "Timed out waiting for return of connection"
                    raise ConnectionUnavailableError(msg) from None
                # Not fully used, so create a new connection.
                conn = self._create_connection()
                # print("created another connection")

                # Connection should be pre-locked for use (avoids timer race).
                assert conn.in_use.locked()

            else:
                # Got unused connection from pool, so lock for use.
                conn.in_use.acquire()

                # Check the connection wasn't closed by the timer.
                if conn.closed:
                    return self._get_connection(
                        timeout=self._time_remaining(timeout, started)
                    )

                # Check the connection is actually usable.
                if self.pre_ping:
                    try:
                        conn.cursor().execute("SELECT 1")
                    except Exception:
                        # Probably connection is closed on server,
                        # but just try to make sure it is closed.
                        conn.close()

                        # Try again to get a connection.
                        return self._get_connection(
                            timeout=self._time_remaining(timeout, started)
                        )

            # Track the connection is now being used.
            self._in_use[id(conn)] = conn

            # Return the connection.
            return conn

    def put_connection(self, conn: TConnection) -> None:
        """Returns connections to the pool, or closes connection
        if the pool is full.

        Unlocks write lock after writer has returned, and
        updates count of readers when readers are returned.

        Notifies waiters when connections have been returned,
        and when there are no longer any readers.
        """
        # Start forgetting if this connection was for reading or writing.
        is_writer, conn.is_writer = conn.is_writer, None

        # Get a lock on tracking usage of connections.
        with self._put_condition:
            # Make sure we aren't dealing with a closed pool
            if self._closed:
                msg = "Pool is closed"
                raise ConnectionPoolClosedError(msg)

            # Make sure we are dealing with a connection from this pool.
            try:
                del self._in_use[id(conn)]
            except KeyError:
                msg = "Connection not in use in this pool"
                raise ConnectionNotFromPoolError(msg) from None

            if not conn.closed:
                # Put open connection in pool if not full.
                if not conn.closing and not self._is_pool_full:
                    self._pool.append(conn)
                    # Close open connection if the pool is full or timer has fired.
                else:
                    # Otherwise, close the connection.
                    conn.close()

            # Unlock the connection for subsequent use (and for closing by the timer).
            conn.in_use.release()

            # If the connection was for writing, unlock the writer lock.
            if is_writer is True:
                self._num_writers -= 1
                self._writer_lock.release()

            # Or if it was for reading, decrement the number of readers.
            elif is_writer is False:
                with self._no_readers:
                    self._num_readers -= 1
                    if self._num_readers == 0 and self._mutually_exclusive_read_write:
                        self._no_readers.notify()

            # Notify a thread that is waiting for a connection to be returned.
            self._put_condition.notify()

    @abstractmethod
    def _create_connection(self) -> TConnection:
        """Create a new connection.

        Subclasses should implement this method by
        creating a database connection of the type
        being pooled.
        """

    def close(self) -> None:
        """Close the connection pool."""
        with self._put_condition:
            if self._closed:
                return
            for conn in self._in_use.values():
                conn.close()
            while True:
                try:
                    conn = self._pool.popleft()
                except IndexError:  # noqa: PERF203
                    break
                else:
                    conn.close()
            self._closed = True

    @staticmethod
    def _time_remaining(timeout: float, started: float) -> float:
        return max(0.0, timeout + started - time())

    def __del__(self) -> None:
        self.close()


TApplicationRecorder_co = TypeVar(
    "TApplicationRecorder_co", bound=ApplicationRecorder, covariant=True
)


class Subscription(Iterator[Notification], Generic[TApplicationRecorder_co]):
    def __init__(
        self,
        recorder: TApplicationRecorder_co,
        gt: int | None = None,
        topics: Sequence[str] = (),
    ) -> None:
        self._recorder = recorder
        self._last_notification_id = gt
        self._topics = topics
        self._has_been_entered = False
        self._has_been_stopped = False

    def __enter__(self) -> Self:
        if self._has_been_entered:
            msg = "Already entered subscription context manager"
            raise ProgrammingError(msg)
        self._has_been_entered = True
        return self

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        if not self._has_been_entered:
            msg = "Not already entered subscription context manager"
            raise ProgrammingError(msg)
        self.stop()

    def stop(self) -> None:
        """Stops the subscription."""
        self._has_been_stopped = True

    def __iter__(self) -> Self:
        return self

    @abstractmethod
    def __next__(self) -> Notification:
        """Returns the next Notification object in the application sequence."""


class ListenNotifySubscription(Subscription[TApplicationRecorder_co]):
    def __init__(
        self,
        recorder: TApplicationRecorder_co,
        gt: int | None = None,
        topics: Sequence[str] = (),
    ) -> None:
        super().__init__(recorder=recorder, gt=gt, topics=topics)
        self._select_limit = 500
        self._notifications: Sequence[Notification] = []
        self._notifications_index: int = 0
        self._notifications_queue: Queue[Sequence[Notification]] = Queue(maxsize=10)
        self._has_been_notified = Event()
        self._thread_error: BaseException | None = None
        self._pull_thread = Thread(target=self._loop_on_pull)
        self._pull_thread.start()

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        try:
            super().__exit__(*args, **kwargs)
        finally:
            self._pull_thread.join()

    def stop(self) -> None:
        """Stops the subscription."""
        super().stop()
        self._notifications_queue.shutdown(  # pyright: ignore[reportAttributeAccessIssue]
            immediate=True
        )
        self._has_been_notified.set()

    def __next__(self) -> Notification:
        # If necessary, get a new list of notifications from the recorder.
        if (
            self._notifications_index == len(self._notifications)
            and not self._has_been_stopped
        ):
            try:
                self._notifications = self._notifications_queue.get()
            except ShutDown:
                pass
            else:
                self._notifications_queue.task_done()
                self._notifications_index = 0

        # Stop the iteration if subscription has been stopped.
        if self._has_been_stopped:
            # Maybe raise thread error.
            if self._thread_error is not None:
                raise self._thread_error
            raise StopIteration

        # Return a notification from previously obtained list.
        notification = self._notifications[self._notifications_index]
        self._notifications_index += 1
        return notification

    def _loop_on_pull(self) -> None:
        try:
            self._pull()  # Already recorded events.
            while not self._has_been_stopped:
                self._has_been_notified.wait()
                self._pull()  # Newly recorded events.
        except BaseException as e:
            if self._thread_error is None:
                self._thread_error = e
            self.stop()

    def _pull(self) -> None:
        while not self._has_been_stopped:
            self._has_been_notified.clear()
            notifications = self._recorder.select_notifications(
                start=self._last_notification_id or 0,
                limit=self._select_limit,
                topics=self._topics,
                inclusive_of_start=False,
            )
            if len(notifications) > 0:
                try:
                    self._notifications_queue.put(notifications)
                except ShutDown:
                    break
                self._last_notification_id = notifications[-1].id
            if len(notifications) < self._select_limit:
                break
